/*  Filename: TabLang.js
 * $Revision: 1.49 $
 *     $Date: 2010/01/27 16:22:39 $
 *
 * TabLang-related code for TabLang service (global XPCOM object)
 * This file is loaded as subscript from components/TabLangService.js
 * Implementations of TabLang (main extension object), Options and
 * TabLangSwitcher wrapper objects go here.
 * 
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TabLang extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is Andrey Lokot.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrey Lokot <crazy-eye@yandex.ru>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// To access main TabLang object use
// Components.classes["tablang@mozdev.org/tablangservice;1"].getService().wrappedJSObject.tablang

// alert() is a shortcut to show message to the user ( replacement of window.alert() )
// Parameters:
//       str - string with message to show
//   caption - caption of message window (optional, "TabLang" is default)
//       win - window message will popup from (optional, current window is default)
function alert(str, caption, win) {
	Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
		.getService(Components.interfaces.nsIPromptService)
		.alert(win ? win : null, caption ? caption : "TabLang", str);
}

// log() produces message to the sstem console
// Parameters:
//   str - message to write (end of line if added automatically)
// Produced message is in the format (timestamp in seconds, Lng is current system lang):
//   012.345 Lng <message>
// For message to appear both "browser.dom.window.dump.enabled" and
// "extensions.tablang.debug.logging" preferences should exist and be set to true
function log(str) {
	var prefs = Components.classes["@mozilla.org/preferences-service;1"]
					.getService(Components.interfaces.nsIPrefBranch);
	try {
		if (prefs.getBoolPref("browser.dom.window.dump.enabled") &&
				prefs.getBoolPref("extensions.tablang.debug.logging")) {
			var time = (Date.now() % 1e6 + 1e6).toFixed();
			var lang = "";
			try {
				if (prefs.getBoolPref("extensions.tablang.debug.logLangs")) {
					lang = Components.classes["tablang@mozdev.org/tablangswitcher;1"]
							.getService(Components.interfaces.ITabLangSwitcher)
							.GetCurrentLangTag().substring(0, 3) + " ";
				}
			}
			catch (err) {
				// No preference == false
			}
			dump(time.substring(1, 4) + "." + time.substring(4, 7) + " " + 
				lang + str + "\n");
		}
	}
	catch (err) {
		// If one of the prefs doesn't exist we don't want logging
	}
}

// reportError() produces error message to console, JS console and alerts it to the user
// Parameters:
//     str - string with error message
//     win - window message for the user will popup from (optional, current window is default).
//           Set to "false" for silent error report (error will be shown in console only)
// warning - (optional) if true warning icon is displayed
function reportError(str, win, warning) {
	var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
							.getService(Components.interfaces.nsIConsoleService);
	var scriptError = Components.classes["@mozilla.org/scripterror;1"]
							.createInstance(Components.interfaces.nsIScriptError);
	scriptError.init("[TabLang] " + str, "", "", undefined, undefined,
		warning ? scriptError.warningFlag : scriptError.errorFlag,
		"chrome javascript");
	consoleService.logMessage(scriptError);
	str = (warning ? "WARNING " : "ERROR ") + str;
	log(str);
	if (win !== false) {
		alert(str, "TabLang", win);
	}
}

// Wrapper for TabLangSwitcher component. Object operates with variables of (informal) type LANG
// which can be both LANG_TAG and LANG_ID so it makes translation between LANG_TAG and LANG_ID
// transparent for the caller. "switcher" member can be used to access TabLangSwitcher component directly.
// ****************************************************************************************
// *** After adding Session Store support only LANG_TAGs are uses as only they are intended
// *** to persist between browser starts. So current TODO is to eleminate use of LANG_IDs
// *** and probably clean up TabLangSwitcher component interface
// *** Note: LANG_IDs no longer supported in TabLangSwitcher for Linux
// ****************************************************************************************
function SwitcherWrapper() {
	// "switcher" variable is used to access TabLangSwitcher component (through ITabLangSwitcher interface)
	this.switcher = Components.classes["tablang@mozdev.org/tablangswitcher;1"]
						.getService(Components.interfaces.ITabLangSwitcher);
}
SwitcherWrapper.prototype = {
	// "currentLang" variable of type LANG which is used to get and set current system input language
	get currentLang () {
		// "currentLang" getter returns current system LANG. It uses
		// GetCurrentLangId() so actual return value is LANG_ID.
		try {
			return this.switcher.GetCurrentLangTag();
		}
		catch (err) {
			reportError("Exception in currentLang getter\n" + err);
			// Trying to hide failure
			return Components.classes["tablang@mozdev.org/tablangservice;1"]
						.getService().wrappedJSObject.options
						.defaultOverride;
		}
	},
	set currentLang (lang) {
		// "currentLang" setter takes "lang" parameter of type LANG. Uses SetCurrentLangId() 
		// if "lang" is LANG_ID or SetCurrentLangTag() if "lang" is LANG_TAG
		if (lang == undefined || lang == "") {
			// Firefox or extensions may simulate session restore with some stub or hand-made
			// JSON data that of course doesn't have "tablang" item stored. As session restore
			// reuses opened tabs "TabOpen" event doesn't fire and language for the tab
			// gets deleted (nsSessionStore.getTabValue returns "" in this case)
			
			// Possibly due to bug 366705 (https://bugzilla.mozilla.org/show_bug.cgi?id=366705):
			// TabOpen may not fire when user tries to load nonexisting page from the local drive
			// so browser.tablangLang would not be set for this tab. This was fixed in Firefox 3.0b1
			
			// As error is not critical don't show alert here not to spoil user experience
			reportError("Language not defined for current tab", /* silent */ false, /* warning */ true);
			lang = Components.classes["tablang@mozdev.org/tablangservice;1"]
						.getService().wrappedJSObject.options
						.defaultOverride;
		}
		try {
			if (typeof(lang) == "number") {
				throw new Error("LANG_IDs no longer supported");
			}
			else {
				this.switcher.SetCurrentLangTag(lang);
			}
		}
		catch (err) {
			reportError("Exception in currentLang setter\nnewlang=" + lang + "\n" + err);
			// New language wasn't set but we still can try to continue execution
		}
	},
	// forceLangTag() function converts LANG to LANG_TAG
	// Takes "lang" parameter of type LANG and returns corresponding LANG_TAG
	// If "lang" is actually LANG_TAG returns "lang" itself. Otherwise
	// uses LangTagFromId() to convert LANG_ID to LANG_TAG.
	forceLangTag: function wrapper_forceLangTag(lang) {
		if (typeof(lang) == "number") {
			throw new Error("LANG_IDs no longer supported");
		}
		return lang;
	}
}

// Options object is used to interact with extension properties
function Options() {
	// *** initialization helper functions ***
	// getLangList() wraps TabLangSwitcher.ListLangTags() and
	// returns list of LANG_TAGs (array of strings)
	function getLangList() {
		var list = { value: [] };
		var count = { value: list.value.count };
		Components.classes["tablang@mozdev.org/tablangswitcher;1"]
			.getService(Components.interfaces.ITabLangSwitcher)
			.ListLangTags(count, list);
		return list.value;
	}
	// *** properties ***
	// prefs - preferences service instance for extension(!) properties
	this.prefs = Components.classes["@mozilla.org/preferences-service;1"]
					.getService(Components.interfaces.nsIPrefService)
					.getBranch("extensions.tablang.");
	// langList - list of languages available in the running system
	this.langList = getLangList();
}
Options.prototype = {
	// validated() - replaces the language if it is not availabe with defaultOverride
	// Takes "lang" - LANG_TAG to validate.
	// Returns "lang" if it is availabe or defaultOverride otherwise
	validated: function options_validated(lang) {
		return (this.langList.indexOf(lang) != -1) ? lang : this.defaultOverride;
	},
	// *** The following properties are correct(!) LNAG_TAGs of corresponding preferences ***
	// defaultOverride - language used instead of one in preferences if that is not available
	get defaultOverride () { var res = this.prefs.getCharPref("defaultLang"); return res ? res : this.langList[0]; },
	// defaultLang - the language to be used for new tabs (with no parent)
	get defaultLang () { return this.validated(this.prefs.getCharPref("defaultLang")); },
	// urlbarLang - the language to be used when editing URLs in address bar
	get urlbarLang () { return this.validated(this.prefs.getCharPref("urlbarLang")); },
	// searchbarLang - the language to be used when editing string to search for in Web Search Bar (upper-right)
	get searchbarLang () { return this.validated(this.prefs.getCharPref("searchbarLang")); },
	// findbarLang - the language to be used when editing string to search for in Page Findbar (bottom)
	get findbarLang () { return this.validated(this.prefs.getCharPref("findbarLang")); },
	// urlbarLangPerTab - true = set separate language for urlbar depending on which tab gets selected
	//                   false = one language for urlbars of all windows (and tabs)
	get urlbarLangPerTab () { return this.prefs.getBoolPref("urlbarLangPerTab"); }
};

var isSessionStoreEnabled;
if (Components.classes["@mozilla.org/xpcom/version-comparator;1"]
				.getService(Components.interfaces.nsIVersionComparator).
				compare(
					Components.classes["@mozilla.org/xre/app-info;1"]
						.getService(Components.interfaces.nsIXULAppInfo)
						.version,
					"3.1a2"
				) >= 0) {
	// Session Store is always enabled since Firefox 3.1a2
	// See https://bugzilla.mozilla.org/show_bug.cgi?id=449596
	isSessionStoreEnabled = true;
	log("isSessionStoreEnabled = true (Firefox 3.1a2 or later detected)");
}
else {
	try {
		// Checking "browser.sessionstore.enabled" preference during startup
		// (it may be chenged later what will have effect only after restart)
		isSessionStoreEnabled = Components.classes["@mozilla.org/preferences-service;1"]
									.getService(Components.interfaces.nsIPrefBranch)
									.getBoolPref("browser.sessionstore.enabled");
	}
	catch (err) {
		// If we can't get this pref for some reason it's better not to use Session Store
		isSessionStoreEnabled = false;
		log(err);
	}
	log("isSessionStoreEnabled = " + isSessionStoreEnabled);
}

// TabLang is the main extension object. It contains all the
// functions to work with "objects" which language is controlled
// "tablangLang" property should be attached to these objects before
// switching between them. This property returns current system
// language for current (focused) object and stored language for others.
// Use TabLang.attachLang() to attach it.
function TabLang() {
	// TabLangSwitcher wrapper instance
	this.wrapper = new SwitcherWrapper;
	// Options instance
	this.options = new Options;
	// currentObject - currently focused object (at least one extension supports).
	// Current system input language is implied to be the language of this object
	// Don't change it manually. To switch to another object use switchTo()
	this.currentObject = null;
	// lastBrowser - last active (focused) browser: the last object
	// of "browser" type that was (or is) saved in currentObject
	// May be not focused. Is used to find "parent" for new tab
	// Don't change it manually. To switch to another object use switchTo()
	this.lastBrowser = null;
	// sharedLangs - to store the languages that should be the same for multiple
	// input controls (such as urlbars/searchbars in different windows) 
	this.sharedLangs = {
		urlbar: this.options.urlbarLang
	};
}
TabLang.prototype = {
	// attachLang() - adds tablangLang property to the object.
	// Adds additional tablangInternalLang varibable
	// to object to store language.
	// Parameters:
	//    obj - object to attach tablangLang property to
	//   lang - initial (stored) language of the object
	attachLang: function tablang_attachLang(obj, lang) {
		var tablang = this;
		// tablangInternalLang - stored language. It is used when oject is not focused
		obj.tablangInternalLang = lang;
		// tablangLang - property, representing language of the object.
		// If object is focused gets and sets current system language
		// else gets and sets tablangInternalLang
		obj.__defineGetter__("tablangLang", function tablangLang_getter() {
			return (tablang.currentObject == this) ? 
				tablang.wrapper.currentLang : this.tablangInternalLang;
		});
		obj.__defineSetter__("tablangLang", function tablangLang_setter(newlang) {
			if (tablang.currentObject == this) {
				tablang.wrapper.currentLang = newlang;
			}
			// Always saving to tablangInternalLang
			// (in case something will go wrong)
			this.tablangInternalLang = newlang;
		});
	},
	// attachSSProperty() - adds property that interfaces with Session Store
	// Parameters:
	//          tab - tab of the browser to attach the property to
	// propertyName - name of the property to add
	//       SSname - Session Store key
	//         lang - initial language of the object
	attachSSProperty: function tablang_attachSSProperty(tab, propertyName, SSname, lang) {
		var browser = tab.linkedBrowser;
		if (isSessionStoreEnabled) {
			// Attaching getter/setter
			var sessionStore = Components.classes["@mozilla.org/browser/sessionstore;1"]
									.getService(Components.interfaces.nsISessionStore);
			browser.__defineGetter__(propertyName, function SSProperty_getter() {
				return sessionStore.getTabValue(tab, SSname);
			});
			browser.__defineSetter__(propertyName, function SSProperty_setter(newLang) {
				sessionStore.setTabValue(tab, SSname, newLang);
			});
		}
		else {
			// Just saving language in member variable (fallback)
		}
		// Initializing
		browser[propertyName] = lang;
	},
	// switchTo() - switches internal extension focus to another object.
	// Use it instead of manually changing currentObject/lastBrowser
	// Parameters:
	//      newobj - object to switch internal focus to (may be null)
	//   isBrowser - should be true/false depending on if newobj is browser.
	//               lastBrowser value is updated accordingly.
	switchTo: function tablang_switchTo(newobj, isBrowser) {
		if (this.currentObject == newobj) {
			log("  switchTo: resetting!");
			if (isBrowser) {
				// Setting lastBrowser to be 100% sure
				this.lastBrowser = newobj;
			}
			return;
		}
		if (this.currentObject) {
			// If there was previous focused object saving its language
			this.currentObject.tablangInternalLang = this.wrapper.currentLang;
			log("  switchTo: saved " + this.currentObject.tablangInternalLang);
		}
		// Switching internally to new object
		this.currentObject = newobj;
		if (isBrowser) {
			this.lastBrowser = newobj;
		}
		// Loading saved language for selected tab
		if (newobj) {
			this.wrapper.currentLang = newobj.tablangInternalLang;
			log("  switchTo: setting " + newobj.tablangInternalLang);
		}
	},
	// setSimpleHandlers() - sets up event handlers to preserve separate input language
	// in the specified element (it must receive focus/blur events, typically it is html input)
	//         obj - JS object of the element ()
	//   sharedKey - if non-empty string:
	//                    used to group elements with the same sharedKey to have
	//                    shared saved input language (language is stored in tablang.sharedLangs)
	//               if "" or null:
	//                    element has its own input language (which is stored in the object itself)
	// initialLang - initial input language for the element
	//  logObjName - elements name for logging
	setSimpleHandlers: function tablang_setSimpleHandlers(obj, sharedKey, initialLang, logObjName) {
		var tablang = this;
		// langObj - used to store the final evaluated object
		// to save language in (depending on sharedKey)
		var langObj;
		if (sharedKey) {
			// Use shared language storage for elements with the same sharedKey
			if (!this.sharedLangs[sharedKey]) {
				// Create storage object if this element is the first with this sharedKey
				this.sharedLangs[sharedKey] = {};
				this.attachLang(this.sharedLangs[sharedKey], initialLang);
			}
			langObj = this.sharedLangs[sharedKey];
		}
		else {
			// Use separate language for this element
			if (!obj.tablangLang) {
				this.attachLang(obj, initialLang);
			}
			langObj = obj;
		}
		// "focus" event handler
		obj.addEventListener("focus", function tablang_simpleHandler_focus(event) {
			log(logObjName + " focus");
			tablang.switchTo(langObj, false);
		}, false);
		// "blur" event handler
		obj.addEventListener("blur", function tablang_simpleHandler_blur(event) {
			log(logObjName + " blur");
			if (tablang.currentObject == langObj) {
				tablang.switchTo(null, false);
			}
		}, false);
	},
	// initTabLang() sets event handlers for all
	// events in this window extension supports
	// Parameters:
	//   win - window to set event handlers for
	initTabLang: function tablang_initTabLang(win) {
		log("New window (DOMContentLoaded)");
		var tablang = this;
		// Focus/blur in tabs
		win.gBrowser.addEventListener(
			"focus",
			function tablang_gBrowser_focus(event) {
				var browser = null;
				if ("nodeName" in event.originalTarget && event.originalTarget.nodeName == "#document"
						&& event.originalTarget.defaultView) {
					// Switching to the browser when it gets focus
					browser = win.gBrowser.getBrowserForDocument(
						event.originalTarget.defaultView.top.document);
				}
				else if ("nodeName" in event.originalTarget && (event.originalTarget.nodeName == "tab"
						|| event.originalTarget.nodeName == "xul:tab") ) {
					// Switching also if the tab itself gets focus
					browser = event.originalTarget.linkedBrowser;
				}
				if (browser) {
					log("focus " + browser.currentURI.spec);
					tablang.switchTo(browser, true);
				}
			},
			true);
		win.gBrowser.addEventListener(
			"blur",
			function tablang_gBrowser_blur(event) {
				var browser = null;
				if ("nodeName" in event.originalTarget && event.originalTarget.nodeName == "#document"
						&& event.originalTarget.defaultView) {
					// Saving language when focus leaves the browser (actually one of its documents,
					// maybe subframes - then focus event for another document in the browser to follow)
					browser = win.gBrowser.getBrowserForDocument(
						event.originalTarget.defaultView.top.document);
				}
				else if ("nodeName" in event.originalTarget && (event.originalTarget.nodeName == "tab"
						|| event.originalTarget.nodeName == "xul:tab") ) {
					// Saving language if the tab itself ws focused (as we switch in this case too)
					browser = event.originalTarget.linkedBrowser;
				}
				if (browser) {
					log("blur " + browser.currentURI.spec);
					tablang.switchTo(null, false);
				}
			},
			true);
		// Sidebar
		var sidebarBrowser = win.document.getElementById("sidebar");
		this.attachLang(sidebarBrowser, this.options.defaultLang);
		sidebarBrowser.addEventListener(
			"focus",
			function tablang_sidebar_focus(event) {
				if ("nodeName" in event.originalTarget && event.originalTarget.nodeName == "#document") {
					// Switching to the browser when it gets focus
					log("focus [sidebar]");
					tablang.switchTo(sidebarBrowser, true);
				}
			},
			true);
		sidebarBrowser.addEventListener(
			"blur",
			function tablang_sidebar_blur(event) {
				if ("nodeName" in event.originalTarget && event.originalTarget.nodeName == "#document") {
					// Saving language when focus leaves the browser (actually one of its documents,
					// maybe subframes - then focus event for another document in the browser to follow)
					log("blur [sidebar]");
					tablang.switchTo(null, false);
				}
			},
			true);
		// Tab Open/Close
		win.gBrowser.tabContainer.addEventListener(
			"TabOpen",
			function tablag_TabOpen(event) {
				tablang.onNewTab(win, event.target);
			},
			false);
		win.gBrowser.tabContainer.addEventListener(
			"TabClose",
			function tablag_TabClose(event) {
				tablang.onTabClose(win, event.target.linkedBrowser);
			},
			false);
		// Support for moving tabs between windows (both drag-n-drop and tearing off).
		// Patching tabbrowser.swapBrowsersAndCloseOther to copy tablangLang.
		if (win.gBrowser.swapBrowsersAndCloseOther) {
			var funcSrc = win.gBrowser.swapBrowsersAndCloseOther.toSource();
			funcSrc = funcSrc.replace('{', '{\
				aOurTab.linkedBrowser.tablangLang = aOtherTab.linkedBrowser.tablangLang;\
				if (aOurTab.linkedBrowser.tablangSavedState) {\
					Components.classes["tablang@mozdev.org/tablangservice;1"]\
						.getService().wrappedJSObject.utils\
						.log("  delete tablangSavedState");\
					delete aOurTab.linkedBrowser.tablangSavedState;\
				}\
				Components.classes["tablang@mozdev.org/tablangservice;1"]\
					.getService().wrappedJSObject.utils\
					.log("swapBrowsersAndCloseOther lang=" + aOurTab.linkedBrowser.tablangLang);\n');
			with (win) eval("win.gBrowser.swapBrowsersAndCloseOther = " + funcSrc);
		}
		// Support for duplicating tabs.
		// Patching tabbrowser.duplicateTab to copy language correctly.
		var old_duplicateTab = win.gBrowser.duplicateTab;
		var bug465215 = Components.classes["@mozilla.org/xpcom/version-comparator;1"]
							.getService(Components.interfaces.nsIVersionComparator).
							compare(Components.classes["@mozilla.org/xre/app-info;1"]
							.getService(Components.interfaces.nsIXULAppInfo).version, "3.1b3") < 0;
		// Note: eval-style patching is not easy here
		// because value returned by original function is required
		win.gBrowser.duplicateTab = function tablang_duplicateTab(aTab) {
			var newTab = old_duplicateTab.call(this, aTab);
			var sessionStore = Components.classes["@mozilla.org/browser/sessionstore;1"]
									.getService(Components.interfaces.nsISessionStore);
			if (bug465215) {
				// Fix for bug 465215: "Session Store uses one variable to store
				// custom data for duplicated tabs (data becomes synchronized)"
				// http://forums.mozillazine.org/viewtopic.php?f=19&t=950705
				// https://bugzilla.mozilla.org/show_bug.cgi?id=465215
				sessionStore.setTabState(newTab, sessionStore.getTabState(newTab));
			}
			// Copying language from tablangLang not tablangInternalLang in case "aTab" is selected tab
			sessionStore.setTabValue(newTab, "tablang", aTab.linkedBrowser.tablangLang);
			return newTab;
		};
		win.addEventListener(
			"unload",
			function tablang_windowUnload(event) { tablang.onWindowClose(win); },
			false);
		// Urlbar
		var urlbar = win.document.getElementById("urlbar").inputField;
		urlbar.addEventListener(
			"focus",
			function tablang_urlbar_focus(event) {
				var browser = win.gBrowser.selectedBrowser;
				log("focus urlbar " + (browser.currentURI ? browser.currentURI.spec : "(new tab)"));
				tablang.switchTo(browser.tablangUrlbarObject, false);
			},
			false);
		urlbar.addEventListener(
			"blur",
			function tablang_urlbar_blur(event) {
				if (tablang.currentObject && "urlbarBrowser" in tablang.currentObject) {
					var browser = tablang.currentObject.urlbarBrowser;
					log("blur urlbar " + (browser.currentURI ? browser.currentURI.spec : "(new tab)"));
					tablang.switchTo(null, false);
				}
			},
			false);
		win.gBrowser.tabContainer.addEventListener(
			"select",
			function(event) {
				if (tablang.currentObject && "urlbarBrowser" in tablang.currentObject) {
					// If switching tabs when urlbar is focused
					// switching to urlbar object for another tab
					var browser = win.gBrowser.selectedBrowser;
					log("focus urlbar " + (browser.currentURI ? browser.currentURI.spec : "(new tab)"));
					tablang.switchTo(browser.tablangUrlbarObject, false);
				}
			},
			false);
		// Web Search bar (upper-right)
		this.setSimpleHandlers(win.document.getAnonymousElementByAttribute(
			win.document.getElementById("searchbar"), "anonid", "searchbar-textbox").inputField,
			null, this.options.searchbarLang, "Searchbar");
		// Page Find textbox (bottom)
		var findbarTextbox = win.document.getAnonymousElementByAttribute(
			win.document.getElementById("FindToolbar"), "anonid", "findbar-textbox");
		if (/* Firefox 2.0 */ !findbarTextbox) findbarTextbox = win.document.getElementById("find-field");
		this.setSimpleHandlers(findbarTextbox.inputField, null, this.options.findbarLang, "Findbar");
	},
	// *** The following functions are event handlers. They may take these input parameters:
	// ***       win - window where event occured
	// ***   browser - browser object for which the event was sent (for tab events)
	// onWindowLoad() is called when window finishes loading
	onWindowLoad: function tabang_onWindowLoad(win) {
		log("onWindowLoad title=" + win.document.title);
		var tablang = this;
		win.setTimeout(function tabang_onWindowLoad_afterTimeout1() {
			win.setTimeout(function tabang_onWindowLoad_afterTimeout2() {
				log("onWindowLoad.(after 2x timeout)");
				// 1. On Linux kxkb from KDE4 (or, possibly, other keyboard switcher) may change
				// keyboard layout for new window if it is used in per-application mode.
				// 2. "Restore closed window" function loads its pages reusing existing tabs so we should
				// set language to one which was restored (note: we have to use setTimeout twice)
				if (tablang.currentObject == win.gBrowser.selectedBrowser) {
					log("onWindowLoad.(after timeout) resetting lang to "
						+ tablang.currentObject.tablangInternalLang);
					tablang.currentObject.tablangLang = tablang.currentObject.tablangInternalLang;
				}
			}, 0);
		}, 0);
	},
	// onNewTab() is called when new tab is created. 
	// Initializes new tab and attaches TabLangState to it
	// Parameter "tab" is a tab object for which the event
	// was sent (instead of "browser" parameter!)
	onNewTab: function tablang_onNewTab(win, tab) {
		log("onNewTab");
		var tablang = this;
		var browser = tab.linkedBrowser;
		// If language was already set for current tab from somewhere
		// (for example, while swapping browsers) we don't need to do anything
		if (browser.tablangLang) {
			log("  skipping (language already set: " + browser.tablangLang + ")");
			log("  !!! Is this triggered anywhere?");
			if (browser.tablangSavedState) {
				log("  delete tablangSavedState");
				delete browser.tablangSavedState;
			}
			return;
		}
		log("onNewTab: attaching lang=" + this.options.defaultLang);
		// Session Store value for language of this tab
		this.attachSSProperty(tab, "tablangInternalLang", "tablang", this.options.defaultLang);
		// Calling attachLang() to attach getter/setter
		// for tablangLang as it were a variable member
		this.attachLang(browser, this.options.defaultLang);
		// Session Store value for language of urlbar when this tab is selected
		this.attachSSProperty(tab, "tablangUrlbarLang", "tablangUrl", this.options.urlbarLang);
		browser.tablangUrlbarObject = {
			urlbarBrowser: browser,
			get tablangInternalLang() {
				// Selecting per-tab/shared language according to current preferece
				return tablang.options.urlbarLangPerTab ? browser.tablangUrlbarLang
					: tablang.sharedLangs.urlbar;
			},
			set tablangInternalLang(newLang) {
				// Saving both per-tab language and shared language
				browser.tablangUrlbarLang = newLang;
				tablang.sharedLangs.urlbar = newLang;
			}
		};
		this.attachLang(browser.tablangUrlbarObject,
			tablang.options.urlbarLangPerTab ? this.options.urlbarLang : tablang.sharedLangs.urlbar);
		// If tab is restored (undoCloseTab) or duplicated
		// (duplicateTab) SessionStore value would be set later
		// so we update current system language if needed
		var options = this.options;
		win.setTimeout(function tablang_onNewTab_afterTimeout() {
			var prevLang = browser.tablangLang;
			browser.tablangLang = browser.tablangInternalLang;
			log("onNewTab.(after timeout) tab: " + ( (prevLang == browser.tablangInternalLang) ? prevLang :
				prevLang + " -> " + browser.tablangInternalLang) );
			if (options.urlbarLangPerTab && win.gBrowser.selectedBrowser == browser) {
				prevLang = browser.tablangUrlbarObject.tablangLang;
				browser.tablangUrlbarObject.tablangLang = browser.tablangUrlbarLang;
				log("onNewTab.(after timeout) url: " + ( (prevLang == browser.tablangUrlbarObject.tablangLang)
					? prevLang : prevLang + " -> " + browser.tablangUrlbarObject.tablangLang) );
			}
		}, 0);
		// Attach information about browser state on the moment tab
		// was created (tablangSavedState) to be used for parent search
		if (this.lastBrowser && this.lastBrowser.currentURI) {
			browser.tablangSavedState = {
				browser: this.lastBrowser,
				uri: this.lastBrowser.currentURI, 
				lang: this.lastBrowser.tablangLang
			};
			log("  added tablangSavedState url=" + browser.tablangSavedState.uri.spec
				+ " lang=" + browser.tablangSavedState.lang);
		}
		// Adding WebProgress listener to detect when URL changes in the current tab
		// so we can try to use document.referrer to detect parent tab
		var tablang = this;
		browser.addProgressListener(
			{
				QueryInterface: function progressListener_QueryInterface(aIID) {
					if (aIID.equals(Components.interfaces.nsIWebProgressListener)
							|| aIID.equals(Components.interfaces.nsISupportsWeakReference)
							|| aIID.equals(Components.interfaces.nsISupports))
						return this;
					throw Components.results.NS_NOINTERFACE;
				},
				onStateChange: function progressListener_onStateChange() {},
				onProgressChange: function progressListener_onProgressChange() {},
				onLocationChange: function progressListener_onLocationChange(aProgress, aRequest, aURI) {
					// If the page in the browser is about:blank possibly tab was
					// just created so we wait for the actual page to be loaded
					// NOTE: we miss the rare case when webpage has link to about:blank 
					if (browser.currentURI.spec != "about:blank") {
						tablang.findParent(win, browser);
						browser.removeProgressListener(this);
					}
				},
				onStatusChange: function progressListener_onStatusChange() {},
				onSecurityChange: function progressListener_onSecurityChange() {}
			},
			Components.interfaces.nsIWebProgress.NOTIFY_LOCATION
		);
	},
	// findParent() uses window.opener and document.referrer
	// to find parent for the new tab
	findParent: function tablang_findParent(win, browser) {
		log("findParent " + browser.currentURI.spec);
		if (!browser.tablangSavedState) {
			log("  no browser.tablangSavedState");
			return;
		}
		if (browser.contentWindow.opener) {
			// If window.opener is defined use it as parent window
			var openerBrowser = null;
			var doc = browser.contentWindow.opener.top ?
				browser.contentWindow.opener.top.document : browser.contentWindow.opener.document;
			var winEnum = Components.classes["@mozilla.org/appshell/window-mediator;1"]
								.getService(Components.interfaces.nsIWindowMediator)
								.getEnumerator("navigator:browser");
			while (!openerBrowser && winEnum.hasMoreElements()) {
				var enumWindow = winEnum.getNext();
				if (enumWindow.gBrowser) {
					openerBrowser = enumWindow.gBrowser.getBrowserForDocument(doc);
				}
			}
			if (openerBrowser) {
				if (openerBrowser == browser.tablangSavedState.browser) {
					// If window.opener is the browser saved in tablangSavedState
					// using the language it had when new tab was created
					browser.tablangLang = browser.tablangSavedState.lang;
				}
				else {
					// Otherwise use current state of window.opener
					reportError("opener != tablangSavedState.browser=" +
						browser.tablangSavedState.uri.spec, /* silent */ false, /* warning */ true);
					if (openerBrowser.tablangLang) {
						browser.tablangLang = openerBrowser.tablangLang;
					}
					else {
						reportError("openerBrowser.tablangLang undefined",
							/* silent */ false, /* warning */ true);
					}
				}
				log("  parent+ opener=" +openerBrowser.currentURI.spec
					+ " lang=" + browser.tablangLang);
			}
		}
		else if (browser.contentDocument.referrer) {
			// Trying to find parent window using document.referrer
			var referrerURI = Components.classes["@mozilla.org/network/io-service;1"]
							.getService(Components.interfaces.nsIIOService)
							.newURI(browser.contentDocument.referrer, null, null);
			if (browser.tablangSavedState.uri.equals(referrerURI)) {
				browser.tablangLang = browser.tablangSavedState.lang;
				log("  parent+ referrer=" + browser.tablangSavedState.uri.spec +
					" lang=" + browser.tablangLang);
			}
		}
		// Delete info saved when tab was opened as it is useless now and
		// we don't want to hold references to browser objects for closed tabs
		log("  delete tablangSavedState");
		delete browser.tablangSavedState;
	},
	// onTabClose() is called when tab is closed to prevent
	// variables from pointing to incorrect browser object
	onTabClose: function tablang_onTabClose(win, browser) {
		log("onTabClose " + browser.currentURI.spec);
		if (this.currentObject == browser) {
			// If current browser is in the closing tab switching
			// to "null" browser to eleminate incorrect links
			this.switchTo(null, true);
		}
		else if (this.lastBrowser == browser) {
			// If the closing tab is just "lastBrowser"
			// no special tricks with language are needed
			this.lastBrowser = null;
		}
	},
	// onWindowClose() is called when window is closed to prevent
	// variables from pointing to incorrect browser object
	onWindowClose: function tablang_onWindowClose(win) {
		log("onWindowClose title=" + win.document.title);
		if (this.currentObject && "contentDocument" in this.currentObject &&
				win.gBrowser.getBrowserForDocument(this.currentObject.contentDocument)) {
			// If current browser is in the closing window switching
			// to "null" browser to eleminate incorrect links
			this.switchTo(null, true);
		}
		else if (this.lastBrowser && "contentDocument" in this.lastBrowser &&
				win.gBrowser.getBrowserForDocument(this.lastBrowser.contentDocument)) {
			// If "lastBrowser" is in the closing window switching
			// to "null" browser to eleminate incorrect links
			this.lastBrowser = null;
		}
	}
};

// initTabLangService() is called from components/TabLangService.js
// It should populate stub service object "obj" with TabLang-related members
function initTabLangService(obj) {
	// tablang - global TabLang instance
	obj.tablang = new TabLang;
	// switcherWrapper - shortcut to tablang.wrapper
	obj.switcherWrapper = obj.tablang.wrapper;
	// options - shortcut to tablang.options
	obj.options = obj.tablang.options;
	// Exporting auxiliary function for external use
	obj.utils = {
		alert: alert,
		reportError: reportError,
		log: log
	}
	// *** DEBUG ***
	// debugReInit() restarts extension. Use for debug only!
	// All browser windows should be closed before calling it.
	// Is used in TabLang tests.
	obj.debugReInit = function debugReInit() {
		this.switcherWrapper = null;
		this.options = null;
		this.tablang = new TabLang;
		this.switcherWrapper = this.tablang.wrapper;
		this.options = this.tablang.options;
	};
}
